module LURank

const CMatrix = Matrix{Cdouble}

default_rho() = 2.0
default_tol(A) = maximum(size(A)) * eps(norm(A[:],Inf))

"""
    lurank

(r,ii,jj,nupdate,t) = lurank(A, rho=default_rho(), tol=default_tol())
determines the numerical rank r of A and a nonsingular submatrix A[ii,jj].
nupdate and t are the number of pivot operations performed and the run time
of the C library.
"""
function lurank(A::CMatrix, rho=default_rho(), tol=default_tol(A))
    m,n = size(A)
    basis = Array{Int}(m)
    p_nupdate = Ref{Int}(0)
    tic()
    r = ccall((:lurank, "liblurank.so"), Int,
              (Int, Int, Ptr{Cdouble}, Cdouble, Cdouble, Ptr{Int}, Ptr{Int}),
              m, n, A, rho, tol, basis, p_nupdate)
    t = toq()
    if r < 0
        msg = @sprintf("lurank error code: %d", r)
        error(msg)
    end
    nupdate = p_nupdate[]
    basis[:] += 1
    ii = setdiff(1:m, basis-n)
    jj = basis[basis.<=n]
    return (r,ii,jj,nupdate,t)
end

"""
    lurank_julia

Implementation in Julia code that does the same arithmetic operations
as the C code.
"""
function lurank_julia(A::CMatrix, rho=default_rho(), tol=default_tol(A))
    m,n = size(A)
    tic()

    # initialize data structures
    mn = m+n
    W = transpose([A eye(m)])
    basis = Array{Int}(m)
    basis[:] = n+1:mn
    is_logical = trues(m)
    nupdate = 0

    # choose maximum entry of A as first pivot
    ipivot = 0
    jpivot = 0
    amax = rho*tol
    for i=1:m
        oi = (i-1)*mn           # offset row i
        for j=1:n
            a = abs(W[oi+j])
            if a > amax amax = a; ipivot = i; jpivot = j; end
        end
    end

    while ipivot > 0
        op = (ipivot-1)*mn      # offset pivot row
        pivot = W[op+jpivot]
        # change basis
        basis[ipivot] = jpivot
        is_logical[ipivot] = jpivot > n
        nupdate += 1
        # update pivotal row
        for j=1:mn
            W[op+j] /= pivot
        end
        W[op+jpivot] = 1.0
        # update nonpivotal rows and search for new pivot
        ipivot_inv = 0;   jpivot_inv = 0;   max_inv = rho
        ipivot_vol = 0;   jpivot_vol = 0;   max_vol = rho
        ipivot_schur = 0; jpivot_schur = 0; max_schur = rho
        for i=1:m
            oi = (i-1)*mn       # offset row i
            if i!=ipivot
                multiplier = -W[oi+jpivot]
                for j=1:mn
                    W[oi+j] += multiplier * W[op+j]
                end
                W[oi+jpivot] = 0.0
            end
            if is_logical[i]
                for j=1:n       # structural replaces logical
                    a = abs(W[oi+j]) / tol
                    if a > max_schur
                        max_schur = a
                        ipivot_schur = i
                        jpivot_schur = j
                    end
                end
                for j=n+1:mn    # structural replaces structural
                    a = abs(W[oi+j])
                    if a > max_vol
                        max_vol = a
                        ipivot_vol = i
                        jpivot_vol = j
                    end
                end
            else
                for j=1:n       # structural replaces structural
                    a = abs(W[oi+j])
                    if a > max_vol
                        max_vol = a
                        ipivot_vol = i
                        jpivot_vol = j
                    end
                end
                for j=n+1:mn    # logical replaces structural
                    a = abs(W[oi+j]) * tol
                    if a > max_inv
                        max_inv = a
                        ipivot_inv = i
                        jpivot_inv = j
                    end
                end
            end
        end
        @assert jpivot_inv != jpivot
        @assert jpivot_vol != jpivot
        @assert jpivot_schur != jpivot
        # choose new pivot
        ipivot = 0
        jpivot = 0
        if ipivot_inv > 0
            ipivot = ipivot_inv
            jpivot = jpivot_inv
        elseif ipivot_vol > 0
            ipivot = ipivot_vol
            jpivot = jpivot_vol
        elseif ipivot_schur > 0
            ipivot = ipivot_schur
                jpivot = jpivot_schur
        end
    end

    r = sum(basis.<=n)
    t = toq()
    ii = setdiff(1:m, basis-n)
    jj = basis[basis.<=n]
    return (r,ii,jj,nupdate,t)
end

end
