#ifndef LURANK_H_
#define LURANK_H_

#include <stdint.h>

typedef int64_t Int;

/*
  lurank

  Purpose:

  Given a real matrix A of any shape and rank, determine the numerical rank
  of A and find a square nonsingular submatrix of that dimension. More
  specifically, for rho >= 1 and tol > 0, find a submatrix A11 such that

    ||A/A11||_C <= rho*tol   and   ||inverse(A11)||_C <= rho/tol,

  where A/A11 is the Schur complement of A11 in A and ||.||_C is the maximum
  absolute entry of a matrix. The dimension of A11 is the numerical rank of A
  (say r) and A11 is chosen so that its minimum singular value is not too much
  smaller than the r-th singular value of A.

  The algorith is described in

  L. Schork, J. Gondzio, Technical Report ERGO 18-002 (2018)
  University of Edinburgh, School of Mathematics
  http://www.maths.ed.ac.uk/ERGO/preprints.html


  Arguments:

  m         number of rows of matrix A. m >= 1.
  n         number of columns of matrix A. n >= 1.
  A         matrix A in column major order.
  rho       controls the maximality of the volume of the submatrix. rho >= 1.
            As default choose rho = 2.0.
  tol       controls the rank detection. tol > 0.
            As default choose tol = max(m,n) * DBL_EPSILON * ||A||_C.
  basis     array of length >= m, uninitialized on entry.
            On successful return basis holds m indices from the range [0,n+m-1].
            The submatrix A11 is composed of rows i and columns j of A for which
            j is in basis and n+i is not in basis.
  p_nupdate optional output argument.
            If not NULL, then on succesful return *p_nupdate holds the number of
            pivot operations performed.

  Returns:

  -i       i-th argument had an illegal value (1 <= i <= 6).
  -9       memory allocation failed.
  r >= 0   succesfully terminated. r is the detected rank of A.
*/

Int lurank(Int m, Int n, const double *A, double rho, double tol, Int *basis,
           Int *p_nupdate);


/*
  Fortran wrapper

  Arguments are passed by reference and the return value is stored in flag.
  Fortran integer type must much Int (change typedef above if necessary).
*/

void lurank_f_(const Int *m, const Int *n, const double *A, const double *rho,
               const double *tol, Int *basis, Int *nupdate, Int *flag);

#endif
