/* Copyright (c) 2018 The University of Edinburgh */

#include <stdbool.h>
#include <stdlib.h>
#include <float.h>
#include <math.h>
#include "lurank.h"

static Int done(void *p1, void *p2, Int err)
{
    if (p1) free(p1);
    if (p2) free(p2);
    return err;
}

Int lurank(Int m, Int n, const double *A, double rho, double tol, Int *basis,
           Int *p_nupdate)
{
    const Int mn = m+n;
    double *W = NULL;
    bool *is_logical = NULL;
    Int ipivot = -1, jpivot = -1, nupdate = 0;

    /* Local variables:
     *
     * W stores the m-times-mn matrix Binv * [A I] in row major order. Binv
     *   denotes the inverse of the basis matrix B and I the identity matrix
     *   of dimension m. Note that rows of W correspond to columns of B.
     *
     * is_logical[i] is true if column i of B comes from the rightmost m columns
     *   of [A I].
     *
     * ipivot and jpivot are the row and column index in W of the next pivot
     *   element. A negative value indicates no further pivot operation.
     *
     * nupdate counts the number of pivot operations.
     */

    /* check arguments */
    if (m <= 0) return -1;
    if (n <= 0) return -2;
    if (!A) return -3;
    if (rho < 1) return -4;
    if (tol <= 0) return -5;
    if (!basis) return -6;

    /* make sure that we do not cycle */
    rho = fmax(rho, 1.0 + 100.0 * DBL_EPSILON);

    /* allocate workspace */
    W = malloc(m*mn * sizeof(double));
    is_logical = malloc(m * sizeof(bool));
    if (!W || !is_logical) return done(W, is_logical, -9);

    /*
     * Build A rowwise and append identity. Initialize basis with logical
     * indices. Keep the position (ipivot,jpivot) of the maximum entry of A.
     */
    double amax = rho*tol;
    for (Int i = 0; i < m; i++)
    {
        double *Wi = W + i*mn;
        for (Int j = 0; j < n ; j++)
        {
            Wi[j] = A[i+j*m];
            double a = fabs(Wi[j]);
            if (a > amax) { amax = a; ipivot = i; jpivot = j; }
        }
        for (Int j = n; j < mn; j++) Wi[j] = 0.0;
        Wi[n+i] = 1.0;
        basis[i] = n+i;
        is_logical[i] = true;
    }

    while (ipivot >= 0)
    {
        double *pivotrow = W + ipivot * mn;
        double pivot = pivotrow[jpivot];

        /* change basis */
        basis[ipivot] = jpivot;
        is_logical[ipivot] = jpivot >= n;
        nupdate++;

        /* update pivotal row */
        for (Int j = 0; j < mn; j++) pivotrow[j] /= pivot;
        pivotrow[jpivot] = 1.0;

        /* Update nonpivotal rows and search for new pivot:
         *
         * (ipivot_inv,jpivot_inv) is the position of the maximum entry in W
         *   for replacing a structural basic by a logical nonbasic column.
         *
         * (ipivot_vol,jpivot_vol) is the position of the maximum entry in W
         *   for exchanging either two logical or two structural columns.
         *
         * (ipivot_schur,jpivot_schur) is the position of the maximum entry in W
         *   for replacing a logical basic by a structural nonbasic column.
         */
        Int ipivot_inv = -1, jpivot_inv = -1;
        Int ipivot_vol = -1, jpivot_vol = -1;
        Int ipivot_schur = -1, jpivot_schur = -1;
        double max_vol = rho, max_schur = rho, max_inv = rho;

        for (Int i = 0; i < m; i++)
        {
            double *row = W + i*mn;
            if (i != ipivot)
            {
                double multiplier = -row[jpivot];
                for (Int j = 0; j < mn; j++) row[j] += multiplier * pivotrow[j];
                row[jpivot] = 0.0;
            }

            if (is_logical[i])
            {
                /* structural replaces logical */
                for (Int j = 0; j < n; j++)
                {
                    double a = fabs(row[j]) / tol;
                    if (a > max_schur)
                    {
                        max_schur = a;
                        ipivot_schur = i;
                        jpivot_schur = j;
                    }
                }
                /* logical replaces logical */
                for (Int j = n; j < mn; j++)
                {
                    double a = fabs(row[j]);
                    if (a > max_vol)
                    {
                        max_vol = a;
                        ipivot_vol = i;
                        jpivot_vol = j;
                    }
                }
            }
            else
            {
                /* structural replaces structural */
                for (Int j = 0; j < n; j++)
                {
                    double a = fabs(row[j]);
                    if (a > max_vol)
                    {
                        max_vol = a;
                        ipivot_vol = i;
                        jpivot_vol = j;
                    }
                }
                /* logical replaces structural */
                for (Int j = n; j < mn; j++)
                {
                    double a = fabs(row[j]) * tol;
                    if (a > max_inv)
                    {
                        max_inv = a;
                        ipivot_inv = i;
                        jpivot_inv = j;
                    }
                }
            }
        }

        /* choose new pivot */
        ipivot = -1;
        jpivot = -1;
        if (ipivot_inv >= 0)
        {
            ipivot = ipivot_inv;
            jpivot = jpivot_inv;
        }
        else if (ipivot_vol >= 0)
        {
            ipivot = ipivot_vol;
            jpivot = jpivot_vol;
        }
        else if (ipivot_schur >= 0)
        {
            ipivot = ipivot_schur;
            jpivot = jpivot_schur;
        }
    }

    Int rank = 0;
    for (Int i = 0; i < m; i++) rank += basis[i] < n;
    if (p_nupdate) *p_nupdate = nupdate;
    return done(W, is_logical, rank);
}

void lurank_f_(const Int *m, const Int *n, const double *A, const double *rho,
               const double *tol, Int *basis, Int *nupdate, Int *flag)
{
    *flag = lurank(*m, *n, A, *rho, *tol, basis, nupdate);
}
